-----------------------------------------------------------------------------
-- |
-- Module      :  Data.Integer
-- Maintainer  :  Malcolm Wallace
--
-- A pure-Haskell (non-primitive) definition of the arbitrary-sized
-- 'Integer' type.
--
-----------------------------------------------------------------------------

module Data.Integer
  ( Integer()
  , decimalShift  -- for tests
  , cmpDigits     -- for tests
  , toDouble
  ) where

import           Prelude hiding (Integer)
import qualified Data.Char as Str
--import         Data.Word
import           Data.Maybe (fromMaybe)
import           Data.Ratio ((%))
--import         Debug.Trace
trace :: String -> a -> a
trace _ = id

-- | Repeatedly chop a list into multiple pieces, given a pair of functions:
--   the first performs a single chop into two pieces, and the other massages
--   the remainder before repeating.  The two functions alternate.  For example,
--
--   >  words         = repeatedly (break isSpace) (dropWhile isSpace)
--   >  dotComponents = repeatedly (break ("."`isPrefixOf`)) tail
--   >  piecesSized n = repeatedly (splitAt n) id
--
repeatedly :: ([a] -> ([a],[a])) -- ^ chopping function
           -> ([a] -> [a])       -- ^ trimming function
           -> [a]                -- ^ the string to chop
           -> [[a]]              -- ^ the resultant segments
repeatedly chop trim s =
    reverse $ fst $
    until (\(res,inp)-> null inp)
          (\(res,inp)-> let (a,b)=chop inp in (a:res,trim b))
          ([],s)


type Word    = Int    -- We fake the Word type, to avoid nasty conversions.
type Digits  = [Word] -- sequence of unsigned "digits", least significant first

data Integer = Zero
             | Positive Digits
             | Negative Digits

largestDigit :: Word
largestDigit  = 999999999  -- 10e9 - 1

divisor      :: Word
divisor       = largestDigit + 1

nthPower     :: Word
nthPower      = 9  -- i.e. 10e9 == divisor

instance  Eq Integer  where
    (==) a b = case a`cmpInteger`b of
                 EQ -> True
                 _  -> False
    (/=) a b = case a`cmpInteger`b of
                 EQ -> False
                 _  -> True

instance Ord Integer where
    compare  = cmpInteger
    (<=) a b = case a`cmpInteger`b of
                 LT -> True
                 EQ -> True
                 GT -> False
    (>)  a b = case a`cmpInteger`b of
                 LT -> False
                 EQ -> False
                 GT -> True
    (<)  a b = case a`cmpInteger`b of
                 LT -> True
                 EQ -> False
                 GT -> False
    (>=) a b = case a`cmpInteger`b of
                 LT -> False
                 EQ -> True
                 GT -> True
    max a b | a>b       = a
            | otherwise = b
    min a b | a<b       = a
            | otherwise = b

-----------------------------------------------------------------------------

instance Read Integer where
    readsPrec _ s
           | all (=='0') s = [(Zero,"")]
           | any (\c->not (Str.isDigit c) && c/='-') s
                                     = error "Not an Integer"
           | otherwise
           = let sign   = if head s == '-' then Negative else Positive
                 s'     = if head s == '-' then tail s   else s
                 chunks = repeatedly (splitAt nthPower) id $
                          reverse $
                          dropWhile (=='0') s'
             in [(sign (map (read . reverse) chunks),"")]

instance Show Integer where   -- code from GHC.Num
    showsPrec p n r
        | p > 6 && n < 0 = '(': integerToString n (')':r)
        | otherwise      =      integerToString n r

integerToString :: Integer -> String -> String
integerToString Zero s          = '0':s
integerToString (Positive ds) s =      concat (reverse (digitsToString ds)) ++ s
integerToString (Negative ds) s = '-': concat (reverse (digitsToString ds)) ++ s

digitsToString :: Digits -> [String]
digitsToString []     = []
digitsToString [d]    = [show d]
digitsToString (d:ds) = [show d] ++ replicate (sizePadding d) "0"
                                 ++ digitsToString ds

sizePadding :: Int -> Int
sizePadding n | n < 1 = 8 -- zero already has one digit
sizePadding n | n < 10 = 8
sizePadding n | n < 100 = 7
sizePadding n | n < 1000 = 6
sizePadding n | n < 10000 = 5
sizePadding n | n < 100000 = 4
sizePadding n | n < 1000000 = 3
sizePadding n | n < 10000000 = 2
sizePadding n | n < 100000000 = 1
sizePadding n | n < 1000000000 = 0


instance Num Integer where  -- code from GHC.Num
    (+) = plusInteger
    (-) = subInteger
    (*) = timesInteger

    negate Zero          = Zero
    negate (Positive ds) = Negative ds
    negate (Negative ds) = Positive ds

    fromInteger          = read . show

    abs Zero             = Zero
    abs (Positive ds)    = Positive ds
    abs (Negative ds)    = Positive ds

    signum Zero          = Zero
    signum (Positive ds) = Positive [1]
    signum (Negative ds) = Negative [1]
  
instance  Enum Integer  where  -- code from GHC.Num
    succ x               = x + 1
    pred x               = x - 1
    toEnum n             = intToInteger n

    fromEnum Zero              = 0
    fromEnum (Negative (d:ds)) = negate d  -- trim big ints
    fromEnum (Positive (d:ds)) = d         -- trim big ints

    {-# INLINE enumFrom #-}
    {-# INLINE enumFromTo #-}
    {-# INLINE enumFromThenTo #-}
    enumFrom x             = error "enumFrom: Integer has no upper bound"
    enumFromTo x lim       = enumDeltaToInteger x 1     lim
    enumFromThenTo x y lim = enumDeltaToInteger x (y-x) lim


enumDeltaToInteger :: Integer -> Integer -> Integer -> [Integer]
enumDeltaToInteger x delta lim
  | delta >= 0 = up_list x delta lim
  | otherwise  = dn_list x delta lim

up_list :: Integer -> Integer -> Integer -> [Integer]
up_list x0 delta lim = go (x0 :: Integer)
                    where
                        go x | x > lim   = []
                             | otherwise = x : go (x+delta)
dn_list :: Integer -> Integer -> Integer -> [Integer]
dn_list x0 delta lim = go (x0 :: Integer)
                    where
                        go x | x < lim   = []
                             | otherwise = x : go (x+delta)

instance Real Integer where
    toRational i = (read $ show i) % 1

toDouble :: Integer -> Double
toDouble Zero          = 0.0
toDouble (Positive ds) =          digitsToDouble ds 0
toDouble (Negative ds) = negate $ digitsToDouble ds 0

digitsToDouble :: Digits -> Int -> Double
digitsToDouble [d]    0 = fromIntegral d
digitsToDouble []     _ = 0.0
digitsToDouble (d:ds) n = (fromIntegral d * 10.0^n) + digitsToDouble ds (n+nthPower)

instance Integral Integer where
--  div      = \a b-> fst $ divModInteger a b
--  mod      = \a b-> snd $ divModInteger a b
    quot     = \a b-> fst $ qRemInteger a b
    rem      = \a b-> snd $ qRemInteger a b
--  divMod   = divModInteger
    quotRem  = qRemInteger
    toInteger= read . show
--  toInteger Zero             = 0
--  toInteger (Positive (d:_)) = d          -- truncated
--  toInteger (Negative (d:_)) = negate d   -- truncated

-- | An Int might require either one or two digits.
intToInteger :: Int -> Integer
intToInteger i
    | i==0                     = Zero
    | i>0 && i <= largestDigit = Positive [i]
    | i>0                      = let (q,r) = quotRem i divisor
                                 in  Positive [r,q]
    | i<0                      = let Positive ds = intToInteger (negate i)
                                 in  Negative ds

-- | Addition of signed numbers.
plusInteger :: Integer -> Integer -> Integer
plusInteger Zero x = x
plusInteger x Zero = x
plusInteger   (Positive ds0)   (Positive ds1) = Positive (addDigits ds0 ds1)
plusInteger a@(Positive ds0) b@(Negative ds1) = subInteger a (Positive ds1)
plusInteger a@(Negative ds0) b@(Positive ds1) = subInteger b (Positive ds0)
plusInteger a@(Negative ds0) b@(Negative ds1) = Negative (addDigits ds0 ds1)

-- | Addition of unsigned numbers is implemented digit-by-digit with carry.
addDigits :: Digits -> Digits -> Digits
addDigits a b = add Nothing a b
  where
    add Nothing       []    []            = []
    add (Just carry)  []    []            = [carry]
    add Nothing       digs  []            = digs
    add Nothing       []    digs          = digs
    add (Just c)      []    digs          = add (Just c) digs []
    add (Just c)     (d:ds) []
                     | dc <= largestDigit = dc: ds
                     | otherwise          = let (q,r) = quotRem dc divisor
                                            in  r: add (Just q) ds []
                     where dc = d+c
    add carry        (d:ds) (e:es)
                     | de <= largestDigit = de: add Nothing ds es
                     | otherwise          = let (q,r) = quotRem de divisor
                                            in  r: add (Just q) ds es
                     where de = d+e+fromMaybe 0 carry

subInteger :: Integer -> Integer -> Integer
subInteger x              Zero         = x
subInteger Zero          (Positive dx) = Negative dx
subInteger Zero          (Negative dx) = Positive dx
subInteger (Positive dx) (Negative dy) = Positive $ addDigits dx dy
subInteger (Negative dx) (Positive dy) = Negative $ addDigits dx dy
subInteger (Negative dx) (Negative dy) = subInteger (Positive dy) (Positive dx)
subInteger (Positive dx) (Positive dy) = let (b,ds) = subDigits dx dy
                                         in if b then positive ds
                                                 else negative ds
  where positive [] = Zero
        positive ds = Positive ds
        negative [] = Zero
        negative ds = Negative ds

-- | Unsigned subtraction is implemented digit-by-digit with carry.
--   The boolean part of the result indicates a positive sign (or zero).
subDigits :: Digits -> Digits -> (Bool,Digits)
subDigits a b | (a`cmpDigits`b)==GT  = (True,  sub a b Nothing)
              | (a`cmpDigits`b)==EQ  = (True,  [])
              | otherwise            = (False, sub b a Nothing)
  where
    sub []     []    Nothing      = []
    sub []     []    (Just carry) = error "subInteger: cannot happen"
    sub []     digs  _            = error "subInteger: ordering violated"
    sub digs   []    Nothing      = digs
    sub (d:ds) []    (Just carry) | carry<=d = (d-carry) -: ds
                                  | carry>d  = let d'=d+divisor
                                               in (d'-carry): sub ds [] (Just 1)
    sub (d:ds) (e:es) Nothing     | e<=d     = (d-e) -: sub ds es Nothing
                                  | e>d      = let d'=d+divisor
                                               in (d'-e): sub ds es (Just 1)
    sub (d:ds) (e:es) (Just c)    | e+c<=d   = (d-(e+c)) -: sub ds es Nothing
                                  | e+c>d    = let d'=d+divisor
                                               in (d'-(e+c)): sub ds es (Just 1)

-- smart list constructor, avoiding MSB zero digits
(-:) :: Word -> Digits -> Digits
0 -: [] = []
d -: ds = d:ds

-- | Digit-by-digit comparison on signed Integers.
cmpInteger :: Integer -> Integer -> Ordering
Zero         `cmpInteger` Zero         = EQ
Zero         `cmpInteger` (Positive _) = LT
Zero         `cmpInteger` (Negative _) = GT
(Positive _) `cmpInteger` Zero         = GT
(Positive a) `cmpInteger` (Positive b) = cmpDigits a b
(Positive _) `cmpInteger` (Negative _) = GT
(Negative _) `cmpInteger` Zero         = LT
(Negative _) `cmpInteger` (Positive _) = LT
(Negative a) `cmpInteger` (Negative b) = cmpDigits a b

-- | Comparison of unsigned digit sequences.
cmpDigits :: Digits -> Digits -> Ordering
[]    `cmpDigits` []     = EQ
[]    `cmpDigits` (_:_)  = LT
(_:_) `cmpDigits` []     = GT
(d:ds)`cmpDigits` (e:es) = case ds`cmpDigits`es of
                             EQ    -> d `compare` e
                             other -> other

{-
-- | Multiplication as repeated addition - extremely slow!
timesInteger :: Integer -> Integer -> Integer
timesInteger a b = fst $ until ((==0).snd)
                               (\(n,i)->(n+a,i-1))
                               (a,b-1)
-}

-- | Long multiplication, just like in primary-school.
timesInteger :: Integer -> Integer -> Integer
timesInteger  Zero          _            = Zero
timesInteger  _             Zero         = Zero
timesInteger (Positive ds) (Positive es) = Positive $ multDigits ds es
timesInteger (Positive ds) (Negative es) = Negative $ multDigits ds es
timesInteger (Negative ds) (Positive es) = Negative $ multDigits ds es
timesInteger (Negative ds) (Negative es) = Positive $ multDigits ds es

-- | We temporarily re-base all the digits from 10^9 to 10^3, to allow
--   enough room to multiply within an Int avoiding overflow. After
--   multiplying, we have digits in base 10^6, which we split and merge back
--   down to 10^3 again.  Some digit sequences are then shifted left, before
--   we re-base back to 10^9, and sum the intermediates.
multDigits :: Digits -> Digits -> Digits
multDigits a b = multiply3 (rebase9to3 a) (rebase9to3 b)

-- | Precondition: arguments are already in base 10^3. Result is in base 10^9.
multiply3 [] es = []
multiply3 ds [] = []
multiply3 ds es = case cmpDigits ds es of
                       LT -> mult es ds
                       _  -> mult ds es
  where
    mult ds es = foldl1 addDigits $
                 map (stripSign "multDigits{Integer}" . rebase3to9) $
                 offset 0 $
                 map (\e->rebase6to3 0 $ map (*e) ds) es

    rebase6to3 carry []     = carry -: []
    rebase6to3 carry (d:ds) = let (q,r) = quotRem (d+carry) (10^3)
                              in r: rebase6to3 q ds

    offset n []       = []
    offset n (ds:dss) = (replicate n 0++ds): offset (n+1) dss

-- | Decimal multiplication, by powers of 10 only.
decimalShift :: Integer -> Int -> Integer
decimalShift Zero          _ = Zero
decimalShift (Positive ds) n = Positive $ digitShift ds n 0
decimalShift (Negative ds) n = Negative $ digitShift ds n 0

digitShift :: Digits -> Int -> Int -> Digits
digitShift digs   0 _     = digs
digitShift digs   n carry | n>=nthPower = carry: digitShift digs (n-nthPower) 0
digitShift []     n 0     = []
digitShift []     n carry = [carry]
digitShift (d:ds) n carry = let (q,r) = quotRem d (10^(nthPower-n))
                            in (r*10^n)+carry : digitShift ds n q


-- | Long division is the trickiest one
qRemInteger :: Integer -> Integer -> (Integer,Integer)
qRemInteger _ Zero = error "quotRem{Integer}: divide by zero"
qRemInteger Zero _ = (Zero,Zero)
qRemInteger a b = case (a,b) of
                    (Positive ds, Positive es) -> mapPosPos (qRemDigits ds es)
                    (Negative ds, Positive es) -> mapNegNeg (qRemDigits ds es)
                    (Negative ds, Negative es) -> mapPosNeg (qRemDigits ds es)
                    (Positive ds, Negative es) -> mapNegPos (qRemDigits ds es)
  where
    mapPosPos (a,b) = (positive a, positive b)
    mapPosNeg (a,b) = (positive a, negative b)
    mapNegPos (a,b) = (negative a, positive b)
    mapNegNeg (a,b) = (negative a, negative b)

    positive [] = Zero
    positive [0] = Zero
    positive ds = Positive ds
    negative [] = Zero
    negative [0] = Zero
    negative ds = Negative ds

-- Special-case some common patterns, to save the expense of doing
-- long division.
qRemDigits :: Digits -> Digits -> (Digits,Digits)
qRemDigits ds [1] = (ds, [])
qRemDigits [d] [e] = let (q,r) = d`quotRem`e in ([q],[r])
qRemDigits ds  es  | ds`cmpDigits`es==LT   = ([],ds)
qRemDigits ds  es  = longDivision (rebase9to3 ds) (rebase9to3 es)

-- This version takes arguments in base 10^3 instead of 10^9, so that carried
-- digits will always fit within a single Word.
-- But the results are in base 10^9.
longDivision :: Digits -> Digits -> (Digits,Digits)
longDivision ds es = case es of
                       [e]   -> walk (reverse ds) e            ([],0)
                       [e,f] -> walk (reverse ds) (e+(f*10^3)) ([],0)
                       _     -> reallyLong ds es
                                           (head
                                            $ stripSign "longDivision{Integer}"
                                            $ rebase3to9
                                            $ reverse
                                            $ take 2
                                            $ reverse es)
  where
    walk []     e (quot,rem) = (stripSign "longDivision{Integer}"
                                          $ rebase3to9 quot
                               ,if rem==0 then [] else [rem])
    walk (d:ds) e (quot,rem) = let (q,r) = (rem * 10^3 + d)`quotRem`e
                               in walk ds e (q:quot,r)

-- | Rebase a string of digits in base 10^9 to digits in base 10^3.
rebase9to3 :: Digits -> Digits
rebase9to3 = norm . concatMap split
  where
    norm     = reverse . dropWhile (==0) . reverse
    split d  = let (q0,r0) = quotRem d  (10^3)
                   (q1,r1) = quotRem q0 (10^3)
               in [r0,r1,q1]

-- | Rebase a string of digits in base 10^3 to an Integer in base 10^9
rebase3to9 :: Digits -> Integer
rebase3to9 = rebase 0 
  where
    rebase n []     = 0
    rebase n (d:ds) = (intToInteger d `decimalShift` n)
                      + rebase (n+3) ds

-- | Get back just the digits from an intermediate calculation on Integer.
stripSign :: String -> Integer -> Digits
stripSign s (Positive ds) = ds
stripSign s Zero          = []
stripSign s _             = error (s++": internal error")



-- Really-long division operates on the principle that dividing the
-- first three digits of the numerator by the first two digits of the divisor
-- is sufficiently close to giving the correct first digit of the result.
-- Then, we subtract the (full divisor, times the calculated digit, shifted
-- left), to give the remainder that serves as the numerator for the next
-- step of the solution.
--
-- When it turns out that the (full divisor, times the calculated digit,
-- shifted left) is greater than the numerator, we adjust the calculated
-- digit by subtracting one, then recompute the subtraction, and carry on.
--
-- On some occasions, the first two digits of the divisor are sufficiently
-- close to (and smaller than) the first *two* digits of the numerator, that we
-- can start there instead.
reallyLong :: Digits -> Digits -> Int -> (Digits,Digits)
reallyLong numerator divisor twoDigits =
    let (q,r) = reallyLong' numerator divisor twoDigits
    in (r3to9 (reverse q), r)

-- The auxiliary is recursive, and returns the quotient reversed, in base 10^3,
-- but the remainder in base 10^9.
reallyLong' :: Digits -> Digits -> Int -> (Digits,Digits)
reallyLong' numerator divisor twoDigits =
-- {-
    trace ("reallyLongDivision{Integer}:"
                 ++"\n\tnumerator="++show numerator
                 ++"\n\tdivisor="++show divisor
                 ++"\n\ttwoDigits="++show twoDigits
                 ++"\n\tnumerator `cmpDigits` divisor ="++
                   show (numerator `cmpDigits` divisor)
                 ) $
-- -}
    -- Terminate when the numerator is smaller than the divisor.
    if numerator `cmpDigits` divisor == LT
    then
         trace ("\n\treturning rem==numerator") $
         ([], r3to9 numerator)
    else
    let guesses :: [(Int,Bool)] -- bool is whether 3-digits from numerator
        guesses = let guess3 = mkguess 3
                      guess2 = mkguess 2 in
                  if guess3 == 0 then []
                  else if guess3 <= 999  then [(guess3,   True)
                                              ,(guess3-1, True)
                                              ,(1,        True)]
                  else if guess3 == 1000 then [(guess2, False)]
                  else --let [lo,_hi] = rebase9to3 [guess3] in
                       --(if lo==0 then ((guess3,   True):) else id)
                                      [(guess2,   False)
                                      ,(guess2-1, False)
                                      ,(1,        False)]
    in
-- {-
    trace (       "\n\tmkguess 3="++show (mkguess 3)
                 ++"\n\tmkguess 2="++show (mkguess 2)
                 ++"\n\tguesses="++show guesses
                 ++"\n\tmap try guesses="++show (map try guesses)
          ) $
-- -}

    -- Refine the guess at a quotient, until it satisfies both:
    --     numerator >= (q*divisor)
    --     numerator -  (q*divisor) < divisor
    let (True,three,q,rem') = firstGood (map try guesses)
        firstGood           = head . filter (\(pos,_,_,_)->pos)
    in
-- {-
    trace ("\n\tq="++show q
          ++"\n\trem'="++show rem'
          ++"\n\tthree="++show three
          ) $
-- -}
    if equalSize three
    then trace ("\n\tequalSize, returning q and rem'") $
         ([q], rem')
    else trace ("\trecursive call with numerator = rem' = "++show (rebase9to3 rem')++"\n-------->") $
         let (subQ,r) = reallyLong' (rebase9to3 rem') divisor twoDigits
             expectedSizeOfQuot = length numerator - length divisor
                                  - (if three then 1 else 0)
             zeros = expectedSizeOfQuot - length subQ
         in trace ("\n\tlength numerator="++show (length numerator)
                  ++"\n\tlength divisor="++show (length divisor)
                  ++"\n\tlength rebase9to3 rem'="++show (length (rebase9to3 rem'))
                  ++"\n\texpectedSizeOfQuot="++show expectedSizeOfQuot
                  ++"\n\tsubQ="++show subQ
                  ++"\n\tr="++show r
                  ++"\n\tzeros: n="++show zeros
                  ) $
            if zeros<=0
            then trace ("\nreturning "++show (subQ++[q],r)) $
                 (q:subQ, r)
            else trace ("\nreturning "++show (subQ++replicate zeros 0++[q],r)) $
                 (q:replicate zeros 0++subQ, r)
  where
    mkguess n = fst $
                (head . r3to9 . reverse . take n . reverse $ numerator)
                `quotRem` -- this quotRem :: Int->Int->(Int,Int)
                twoDigits

    lengthDiff3 threeDigits = length numerator - length divisor
                              + (if threeDigits then 0 else 1)
    equalSize   threeDigits = lengthDiff3 threeDigits == 0

    try :: (Int,Bool) -> (Bool,Bool,Int,Digits)
    try (q,threeDigits) =
        let subterm   = rebase9to3 [q] `multiply3` divisor
            (pos,rem) = r3to9 numerator
                        `subDigits`
                        digitShift subterm (max 0 (lengthDiff3 threeDigits-1)*3) 0
        in  (pos,threeDigits,q,rem)

r3to9 :: Digits -> Digits
r3to9 = stripSign "reallyLong{Integer}" . rebase3to9

